#include "main.h"

void simulation_init(int argc, const char* argv[], simulation_mesh_t& sm)
{
    //
    // load tetmesh files
    //
    std::string input_fpath_basename(argc == 1 ? "bunny" : argv[1]); // load the bunny by default
    sm.tetmesh.load(input_fpath_basename);

    //
    // rescale the input tetmesh vertices according to our needs
    //
    const vec3 local_mesh_translation(.0f, .0f, .0f);
    const vec3 local_mesh_scale(1.f, 1.f, 1.f);
    normalize_mesh(sm.tetmesh, local_mesh_translation, local_mesh_scale);

    //
    // set material parameters
    //
    sm.material_parameters.lambda = 100.45f;
    sm.material_parameters.mu = 6000.f;

    //
    // Create the simulation nodes of our finite element mesh
    //
    const int node_count = sm.tetmesh.vertices.size();

    for (int i = 0; i < node_count; ++i) {
        node_t node;
        node.mass = 1.0f;
        node.position = sm.tetmesh.vertices[i];
        node.position_prev = node.position;
        node.force = vec3(0, 0, 0);
        sm.nodes.push_back(node);
    } // for (int i = 0; i < node_count; ++i) {

    //
    // Compute constant variables of each finite element
    //

    const int element_count = sm.tetmesh.tetrahedra.size() / 4;

    for (int i = 0; i < element_count; ++i) {
        const int base_index = i * 4;
        finite_element_t element;

        // 1) TODO compute inverse of undeformed-shape matrix 
        const std::vector<int>& tet = sm.tetmesh.tetrahedra;
        const vec3& X1 = sm.nodes[tet[base_index + 0]].position;
        // ...
        
        // element.undeformed_shape_matrix_inverse = ;

        // 2) TODO undeformed-shape volume of element
        // element.volume = 

        // 3) undeformed-shape normals
        for (int j = 0; j < 4; ++j) { // for each node of finite element

            const int node_vertex_idx = tet[base_index + j]; // global index of the node

            for (int k = 0; k < 3; ++k) { // for each vertex of opposite face
                const int local_idx = (((j + 1) + k) % 4);
                // save vertex of opposite face (note: winding order may be CW or CCW)
                element.undeformed_shape_normals[j].face[k] = sim_mesh.tetmesh.tetrahedra[base_index + local_idx];
            }

            const vec3& X1_ = sm.nodes[element.undeformed_shape_normals[j].face[0]].position;
            const vec3& X2_ = sm.nodes[element.undeformed_shape_normals[j].face[1]].position;
            const vec3& X3_ = sm.nodes[element.undeformed_shape_normals[j].face[2]].position;
            const vec3& X4_ = sm.nodes[node_vertex_idx].position; // node which is opposite to face
            const float area = calculate_face_area(X1_, X3_, X2_);
            const vec3 area_weighted_normal = calculate_face_normal(X1_, X3_, X2_, X4_) * area;

            element.undeformed_shape_normals[j].normal = area_weighted_normal * -1.0f; // save element face normal
        }

        //assert(element.volume > 0);

        sm.finite_elements.push_back(element);

    } // for (int i = 0; i < element_count; ++i) {

    printf("simulation init success!\n");
}

void simulation_step(simulation_mesh_t& sm, const float timestep_size, const float timestep_size_prev)
{
    const int node_count = sm.nodes.size();

    //
    // reset all vertex forces
    //
    for (int i = 0; i < node_count; ++i) {
        sm.nodes[i].force = vec3(0, 0, 0);

        float magnitude = 16.f;
        sm.nodes[i].force = sm.nodes[i].force + vec3(0, magnitude * std::cos(accumulator), /*magnitude * std::sin(accumulator)*/0);
    }

    //
    // compute elastic forces
    //

    const int element_count = sm.finite_elements.size();

    for (int i = 0; i < element_count; ++i) { // for each finite element

        finite_element_t& element = sm.finite_elements[i]; // reference to current element
        const int base_index = i * 4; // the starting index of [the vertex indices which define the current element]

        // TODO: compute DEFORMED-shape element matrix
        const std::vector<int>& tet = sm.tetmesh.tetrahedra;
        const vec3& x1 = sm.nodes[tet[base_index + 0]].position;
        // ...

        // element.deformed_shape_matrix = ;

        // TODO compute DEFORMATION GRADIENT
        // element.deformation_gradient_matrix = ;

        // TODO compute STRAIN (green strain)
        const mat3& F = element.deformation_gradient_matrix;
        const mat4 F_as_mat4 = mat4::FromMat3(F);
        const mat4 F_as_mat4_transposed = mat4::Transpose(F_as_mat4);
        const mat3 F_transpose = mat4::ToMat3(F_as_mat4_transposed);
        // element.strain_matrix = ;

        // TODO compute STRESSES
        const mat3& strain = element.strain_matrix;
        const float lambda = sm.material_parameters.lambda;
        const float mu = sm.material_parameters.mu;
        // element.stress_matrix = ;

        // TODO compute NODAL FORCES
        for (int j = 0; j < 4; ++j) { // for each node of the current element

            const vec3& normal = element.undeformed_shape_normals[j].normal;
            vec3 force_on_face = 0; // TODO: compute the force on the element-face which is opposite to node

            if (fabs(vec3::Length(force_on_face)) < flp_epsilon) { // is the force too small..?
                force_on_face = vec3(0, 0, 0); // just set to zero
            } else { // ... otherwise we have a sufficiently large force, so we can apply it
                const int* const face = element.undeformed_shape_normals[j].face;
                // const vec3 distributed_nodal_force = ; // TODO: calcuclate the force to be distributed on the vertices on the face

                // TODO: add "distributed_nodal_force" to each node defining "face"
                // sm.nodes[face[0]].force = sm.nodes[face[0]].force + distributed_nodal_force;
                // ...
            }
        }
    } // for (int i = 0; i < element_count; ++i) {

    //
    // update vertex positions
    //

    float time_ratio = timestep_size / timestep_size_prev;

    for (int i = 0; i < node_count; ++i) {
        node_t& node = sm.nodes[i];

        // Numerical integration step
        {
            // Example verlet integration: 
            // x(t + dt)    = 2x(t) - x(t - dt) + ddx(t)dt^2
            //              = x(t) + [x(t)  - x(t - dt)] + (f(t)/m)dt^2
            // Velocity does not show up explicitely, instead the position at the previous time step has to be stored

            vec3 velocity = node.position - node.position_prev;
            node.position_prev = node.position;

            if (fabs(vec3::Length(velocity)) < flp_epsilon) {
                velocity = vec3(0, 0, 0);
            }

            const float velocityFalloff = 0.5f * time_ratio;
            const vec3 delta = velocity * velocityFalloff;

            node.position = node.position + delta + ((node.force / node.mass) * (timestep_size * timestep_size));
        }

        node.position = vec3::Clamp(node.position, vec3(collision_min), vec3(collision_max)); // simple collision detection
    }
}

int main(int argc, const char* argv[])
{
    //
    // initialise and setup the OpenGL Utility Toolkit (GLUT)
    //
    const int screen_width = 768;
    const int screen_height = 512;

    glutInit(&argc, (char**)argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
    glutInitWindowSize(screen_width, screen_height);
    glutInitWindowPosition(0, 0);
    glutCreateWindow("CAV MASS-SPRING SYSTEM [BUILD: " __DATE__ "]");
    glutKeyboardFunc(key_pressed_down);
    glutKeyboardUpFunc(key_released_up);
    glutSpecialFunc(special_key_pressed_down);
    glutSpecialUpFunc(special_key_released);
    glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);
    glutReshapeFunc(reshape);
    glutDisplayFunc(render);
    glutIdleFunc(update);

    //
    // setup OpenGL rendering
    //

    glEnable(GL_MULTISAMPLE);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)screen_width / (GLfloat)screen_height, 0.1f, 100.0f);
    glShadeModel(GL_FLAT); // Enables Flat Shading
    glClearDepth(1.0f); // Depth Buffer Setup
    glEnable(GL_DEPTH_TEST); // Enables Depth Testing
    glDepthFunc(GL_LEQUAL);
    glViewport(0, 0, screen_width, screen_height);
    glClearColor(0.1, 0.5, 0.5, 1.0);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Really Nice Perspective Calculations
    glLineWidth(2);

    //
    // initialise simulation
    //

    simulation_init(argc, argv, sim_mesh);

    frame_rate_mean.add(fps_limit);

    printf("Tetrahedrons, Initial iterations, Time step per sec, FPS target, FPS variance, Min FPS, Max step\n");
    printf("%d, %d, %f, %d, %d, %d, %f\n", (int)sim_mesh.tetmesh.tetrahedra.size() / 4, iterations_per_frame, step_per_second, fps_target, fps_variance, fps_min, max_timestep);
    printf("\nTime (secs), Frame Time (ms), Smoothed FPS, FPS, FPS Target, Time step, Iterations\n");

    glutMainLoop();

    return 0;
}
