#ifndef __MAIN_H__
#define __MAIN_H__

#include <algorithm>
#include <cmath>
#include <float.h> // FLT_MIN FLT_MAX
#include <iostream>
#include <stdio.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#include <OpenGL/OpenGL.h>
#else
#include <GL/glut.h>
#endif

#include "mat.h"
#include "tetmesh.h"
#include "vec.h"

#define GLUT_KEY_ESCAPE 27

const float collision_min = -1.08f;
const float collision_max = 1.08f;
const float flp_epsilon = 1e-4f;

int iterations_per_frame = 50;
const int fps_limit = 60;
const int fps_variance = 4;
const int fps_target = 30;
const int fps_min = 24;
const float frame_length = 1000.0 / (float)fps_limit;
const float step_per_second = 1.0f;
const float max_timestep = 0.01f;
const int max_iterations_per_frame = 1000;
float frame_time_last = 0;

float camera_rotation_speed = 0.001f;
float camera_rotation = -90;
float accumulation_speed = 0.003f;
float accumulator = 0;
int seconds_passed = 0;

class moving_mean_t {
public:
    moving_mean_t()
    {
        reset();
    }

    void add(float f)
    {
        sum = sum * 0.95f + f;
        weights = weights * 0.95f + 1.0f;
    }

    float mean()
    {
        return sum / weights;
    }

    void reset()
    {
        sum = 0;
        weights = 0;
    }

private:
    float sum;
    float weights;
} frame_rate_mean;

struct element_face_normal_t {
    vec3 normal;
    int face[3];
};

struct finite_element_t {
    //
    // constant variables
    //
    mat3 undeformed_shape_matrix_inverse;
    element_face_normal_t undeformed_shape_normals[4]; // area weighted normals of the face opposite the node the force is acting upon in the rest configuration.
    float volume;

    //
    // updated variables
    //
    mat3 deformed_shape_matrix; // world space
    mat3 deformation_gradient_matrix;
    mat3 strain_matrix;
    mat3 stress_matrix;
};

struct node_t {
    vec3 position;
    vec3 position_prev;
    vec3 force;
    float mass;
};

struct material_t {
    float lambda;
    float mu;
};

struct simulation_mesh_t {
    tetmesh_t tetmesh; // tetrahedron mesh (as loaded from file).
    material_t material_parameters;
    std::vector<node_t> nodes;
    std::vector<finite_element_t> finite_elements;
} sim_mesh;

// defined in main.cpp
extern "C" void simulation_step(simulation_mesh_t& sm, const float timestep_size, const float timestep_size_prev);

vec3 calculate_face_normal(vec3 const& v1, vec3 const& v2, vec3 const& v3, vec3 const& non_coplanar_point)
{
    vec3 v12 = (v2 - v1);
    vec3 v13 = (v3 - v1);
    vec3 orth = vec3::Cross(v12, v13);
    vec3 n = vec3::normalize(orth);
    vec3 v = vec3::normalize(non_coplanar_point - v1);

    if (vec3::Dot(v, n) > 0) {
        return -n; // must point in opposite dir (i.e. outside element)
    }

    return n;
}

float calculate_face_area(vec3 const& v1, vec3 const& v2, vec3 const& v3)
{
    vec3 v12 = (v2 - v1);
    vec3 v13 = (v3 - v1);

    float term1 = std::pow(v12.y * v13.z - v12.z * v13.y, 2);
    float term2 = std::pow(v12.z * v13.x - v12.x * v13.z, 2);
    float term3 = std::pow(v12.x * v13.y - v12.y * v13.x, 2);

    return std::sqrt(term1 + term2 + term3) / 2.0f;
}

void draw_mesh()
{
    glPushMatrix();
    {
        glBegin(GL_TRIANGLES);
        const int boundary_face_count = (int)sim_mesh.tetmesh.boundaryFaces.size() / 3;
        for (int i = 0; i < boundary_face_count; ++i) { // for each tetrahedron

            int face_1st_vertex_idx = i * 3;
            int face[3];

            for (int j = 0; j < 3; ++j) { // for each vertex of face
                face[j] = sim_mesh.tetmesh.boundaryFaces[face_1st_vertex_idx + j];
            }

            // draw the face
            for (int j = 0; j < 3; ++j) // for each vertex of face
            {
                int vertex_index = face[j];
                const vec3 v = sim_mesh.nodes[vertex_index].position;

                glVertex3f(v.x, v.y, v.z);
            }
        }
        glEnd();
    }
    glPopMatrix();
}

void draw_plane()
{
    glBegin(GL_TRIANGLES);
    glVertex3f(-1, 0, -1); // back left
    glVertex3f(-1, 0, 1); // front left
    glVertex3f(1, 0, 1); // front right

    glVertex3f(1, 0, 1); // front right
    glVertex3f(1, 0, -1); // back right
    glVertex3f(-1, 0, -1); // back left
    glEnd();
}

void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // reset the color and depth buffer

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0f, 0.f, -6.0f);
    glRotatef(15, 1, 0, 0);
    camera_rotation += camera_rotation_speed * frame_time_last;
    glRotatef(camera_rotation, 0, 1, 0);

    //
    // render floor plane
    //
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f(0.0f, 0.0f, 0.0f);
    glPushMatrix();
    {
        glTranslatef(0.f, collision_min, 0.f);
        draw_plane();
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glColor3f(0.75f, 0.75f, 0.75f);
        draw_plane();
    }
    glPopMatrix();

    //
    // render ceiling plane
    //
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f(0.6f, 1.0f, 0.2f);
    glPushMatrix();
    {
        glTranslatef(0.f, collision_max, 0.f);
        draw_plane();
    }
    glPopMatrix();

    //
    // render mesh
    //
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f(0.0f, 0.0f, 0.0f);
    draw_mesh();
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glColor3f(0.0f, 0.0f, 1.0f);
    draw_mesh();

    glutSwapBuffers();
}

void update()
{
    accumulator += accumulation_speed * frame_time_last;

    float timestep_size = step_per_second / std::sqrt((float)iterations_per_frame) / fps_target;
    timestep_size = std::min(timestep_size, max_timestep);
    float timestep_size_prev = timestep_size;
    static int lastPrintTime = 0;
    int frame_rate_target_current = fps_target;
    float frame_time_start_ms = (float)glutGet(GLUT_ELAPSED_TIME);

    glutPostRedisplay(); // tell glut to render a frame

    //
    // run several simulation timesteps
    //
    for (int i = 0; i < iterations_per_frame; ++i) {
        simulation_step(sim_mesh, timestep_size, timestep_size_prev);
    }

    seconds_passed = (int)glutGet(GLUT_ELAPSED_TIME) / 1000;
    float frame_time_elapsed_ms = (float)glutGet(GLUT_ELAPSED_TIME) - frame_time_start_ms;
    float fps = 1000.0f / frame_time_elapsed_ms;
    frame_time_last = frame_time_elapsed_ms;

    if (seconds_passed > lastPrintTime) {
        printf("%d, %f, %f, %f, %d, %f, %d\n", seconds_passed, frame_time_elapsed_ms, frame_rate_mean.mean(), fps, frame_rate_target_current, timestep_size, iterations_per_frame);
        lastPrintTime = seconds_passed;
    }

    frame_rate_mean.add(fps);
    timestep_size_prev = timestep_size;

    if (frame_rate_mean.mean() > frame_rate_target_current + fps_variance && iterations_per_frame < max_iterations_per_frame) {
        iterations_per_frame++;
    } else if (frame_rate_mean.mean() < frame_rate_target_current - fps_variance && iterations_per_frame > 1) {
        iterations_per_frame--;
    }

    timestep_size = step_per_second / std::sqrt((float)iterations_per_frame) / frame_rate_mean.mean();

    if (timestep_size > max_timestep) {
        timestep_size = max_timestep;
        if (frame_rate_target_current > fps_min) {
            frame_rate_target_current--;
        }
    } else if (frame_rate_target_current < fps_target) {
        frame_rate_target_current++;
    }
}

void key_pressed_down(unsigned char key, int x, int y)
{
    switch (key) {
    case GLUT_KEY_ESCAPE:
        std::exit(0);
        break;
    }
}

void key_released_up(unsigned char key, int x, int y)
{
}

void special_key_pressed_down(int key, int x, int y)
{
}

void special_key_released(int key, int x, int y)
{
}

void reshape(int width, int height)
{
    glViewport(0, 0, (GLsizei)width, (GLsizei)height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);
}

#endif
