#include <string>
#include <vector>

#ifdef __APPLE__
#include <GLUT/glut.h>
#include <OpenGL/OpenGL.h>
#else
#include <GL/glut.h>
#endif

#include "tetmesh.h"
#include "vec.h"

struct Particle {
    double mass;
    vec3 pos, vel, frc;
};

struct Spring {
    double r; // rest length
    int i, j;
};

struct SimulationParameters {
    double dt, k_s, k_d;
    int timesteps_per_frame;
};

struct Object {
    std::vector<Particle> particles;
    std::vector<Spring> springs;
};

static const double FEPS = 1e-6;
SimulationParameters params;
Object my_object;
tetmesh_t tetmesh;
float collision_min = -1.87;

void update_logic()
{
    params.dt = 0.002; // ideally you want to set this according to the elapsed frame time

    glutPostRedisplay(); // tell glut to render a frame

    // run the timesteps
    for (int i = 0; i < params.timesteps_per_frame; i++) {

        // TODO: initialize force accumulator, apply gravity

        // TODO: compute and accumulate spring forces

        // TODO: time integration
        
        // ground collision
        for (std::vector<Particle>::iterator p = my_object.particles.begin(); p != my_object.particles.end(); p++) {
            if (p->pos.y < collision_min) {
                p->pos.y = collision_min;
                p->vel.y = 0.0;
            }
        }
    }
}

void draw_plane()
{
    glBegin(GL_TRIANGLES);
    glVertex3f(-1, 0, -1); // back left
    glVertex3f(-1, 0, 1); // front left
    glVertex3f(1, 0, 1); // front right

    glVertex3f(1, 0, 1); // front right
    glVertex3f(1, 0, -1); // back right
    glVertex3f(-1, 0, -1); // back left
    glEnd();
}

void draw()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // reset the color and depth buffer
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0f, 0.f, -6.0f);
    glRotatef(15, 1, 0, 0);

    // draw spring mesh
    glColor3f(0.4, 0.2, 0.9);
    glPushMatrix();
    {
        glBegin(GL_LINES);
        for (int i = 0; i < my_object.springs.size(); ++i) {
            const Spring& s = my_object.springs.at(i);
            const vec3& ipos = my_object.particles[s.i].pos;
            const vec3& jpos = my_object.particles[s.j].pos;
            glVertex3f(ipos.x, ipos.y, ipos.z);
            glVertex3f(jpos.x, jpos.y, jpos.z);
        }
        glEnd();
    }
    glPopMatrix();

    // draw floor
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f(0.0f, 0.0f, 0.0f);
    glPushMatrix();
    {
        glTranslatef(0.f, collision_min, 0.f);
        glScalef(4.f, 4.f, 4.f);
        draw_plane();
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glColor3f(0.75f, 0.75f, 0.75f);
        draw_plane();
    }
    glPopMatrix();

    glutSwapBuffers();
}

int main(int argc, char* argv[])
{
    //
    // initialise and setup the OpenGL Utility Toolkit (GLUT)
    //
    const int screen_width = 768;
    const int screen_height = 512;

    glutInit(&argc, (char**)argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
    glutInitWindowSize(screen_width, screen_height);
    glutInitWindowPosition(0, 0);
    glutCreateWindow("CAV MASS-SPRING SYSTEM [BUILD: " __DATE__ "]");
    glutDisplayFunc(draw);
    glutIdleFunc(update_logic);

    //
    // setup OpenGL rendering
    //

    glEnable(GL_MULTISAMPLE);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)screen_width / (GLfloat)screen_height, 0.1f, 100.0f);
    glShadeModel(GL_FLAT); // Enables Flat Shading
    glClearDepth(1.0f); // Depth Buffer Setup
    glEnable(GL_DEPTH_TEST); // Enables Depth Testing
    glDepthFunc(GL_LEQUAL);
    glViewport(0, 0, screen_width, screen_height);
    glClearColor(0.1, 0.5, 0.5, 1.0);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Really Nice Perspective Calculations
    glLineWidth(2);

    //
    // initialise our simulation
    //
    tetmesh.load("dodecahedron", true);

    //
    // rescale the input tetmesh vertices according to our needs
    //
    const vec3 local_mesh_translation(.0f, 4.0f, .0f);
    const vec3 local_mesh_scale(1.f, 1.f, 1.f);
    normalize_mesh(tetmesh, local_mesh_translation, local_mesh_scale);

    for (int i = 0; i < tetmesh.vertices.size(); ++i) {
        Particle p;
        p.mass = 1;
        p.frc = vec3(0);
        p.pos = tetmesh.vertices[i];
        p.vel = vec3(0);
        my_object.particles.push_back(p);
    }

    for (int i = 0; i < tetmesh.edges.size() / 2; i++) {
        Spring s;
        s.i = tetmesh.edges[i * 2 + 0];
        s.j = tetmesh.edges[i * 2 + 1];
        s.r = vec3::Length(my_object.particles[s.i].pos - my_object.particles[s.j].pos);

        my_object.springs.push_back(s);
    }

    params.timesteps_per_frame = 16;
    params.k_s = 1e4;
    params.k_d = 0.05;

    printf("params.timesteps_per_frame=%d params.k_s=%f params.k_d=%f\n", params.timesteps_per_frame, params.k_s, params.k_d);

    glutMainLoop();
}